#ifndef ML_SVM_HPP
#define ML_SVM_HPP
#include <memory>
#include <vector>

namespace image{

namespace ml{

#include <libsvm/svm.h>

template<typename attribute_type,typename classification_type>
class svm
{
    unsigned int attribute_dimension;
	std::vector<double> y_buf;
    std::vector<std::vector<svm_node> > data;
    std::vector<svm_node*> x_buf;
	svm_parameter param;
	svm_problem prob;
	svm_model* model;
public:
	svm(void):model(0)
    {
        // default values
        param.svm_type = C_SVC;
        param.kernel_type = RBF;
        param.degree = 3;
        param.gamma = 0;	// 1/num_features
        param.coef0 = 0;
        param.nu = 0.5;
        param.cache_size = 100;
        param.C = 100;
        param.eps = 1e-3;
        param.p = 0.1;
        param.shrinking = 1;
        param.probability = 0;
        param.nr_weight = 0;
        param.weight_label = NULL;
        param.weight = NULL;
    }
	~svm(void)
	{
        svm_free_and_destroy_model(&model);
	}
public:
    template<typename attributes_iterator_type,typename classifications_iterator_type>
    void learn(attributes_iterator_type attributes_from,
               attributes_iterator_type attributes_to,
               size_t attribute_dimension_,
               classifications_iterator_type classifications_from)
    {
        // put data in prob
        attribute_dimension = attribute_dimension_;
        {
			prob.l = attributes_to-attributes_from;
            y_buf.resize(prob.l);
			x_buf.resize(prob.l);
			data.resize(prob.l);
            for (size_t index = 0;index < prob.l;++index)
            {
                data[index].resize(attribute_dimension+1); // libsvm has -1 index at the end
                for(unsigned int j = 0;j < attribute_dimension;++j)
                {
                    data[index][j].index = j+1;
                    data[index][j].value = attributes_from[index][j];
                }
                data[index].back().index = -1;// libsvm has -1 index at the end
                x_buf[index] = &*(data[index].begin());
                y_buf[index] = classifications_from[index];
			}
			prob.y = &*y_buf.begin();
            prob.x = &*x_buf.begin();
        }
        param.gamma = 1.0/(double)attribute_dimension;
		if(model)
            svm_free_and_destroy_model(&model);
		model = svm_train(&prob,&param);
    }
    template<typename sample_iterator_type>
    classification_type predict(sample_iterator_type predict_attributes_) const
    {
        std::vector<svm_node> x(attribute_dimension+1);
        for(unsigned int j = 0;j < attribute_dimension;++j)
        {
            x[j].index = j+1;
            x[j].value = predict_attributes_[j];
        }
        x.back().index = -1;// libsvm has -1 index at the end
        return svm_predict(model,&x[0]);
    }
};


}// ml

}// image


#endif//ML_SYM_HPP
